import codecs
from pathlib import Path
from typing import Optional

import httpx

from acquisition_statute_parser import parse

from .sanitizer import statute_sanitizer


def sanitize_content(raw: str):
    """Product uniform html tags

    Args:
        raw (str): Raw text fetched

    Returns:
        [type]: Raw text cleaned
    """
    return statute_sanitizer.sanitize(raw)


def raw_from_url(idx: int, statute_type: Optional[int]) -> str:
    """The supplied `idx` completes the url

    Args:
        idx (int): [description]
        statute_type (Optional[int]): If not supplied, the default is 2 for RAs

    Returns:
        str: [description]
    """
    if not statute_type:
        statute_type = 2
    base = f"https://elibrary.judiciary.gov.ph/thebookshelf/showdocsfriendly"
    url = f"{base}/{statute_type}/{idx}"
    response = httpx.get(url, verify=False)
    return response.text


def get_context(statute_type: int) -> str:
    """Peculiar mapping of URL string to human readable context, used as a folder

    Args:
        statute_type (int): A number which maps to a URL path

    Returns:
        str: The local folder to use in context
    """
    if statute_type == 2:
        folder = "ra"

    elif statute_type == 3:
        folder = "const"

    elif statute_type == 26:
        folder = "pd"

    elif statute_type == 5:
        folder = "eo"

    elif statute_type == 25:
        folder = "bp"

    elif statute_type == 29:
        folder = "ca"

    elif statute_type == 28:
        folder = "act"

    return folder


def raw_to_file(idx: int, statute_type: int = 2) -> None:
    """Create an html file in a local directory determined by the parameters

    Args:
        idx (int): The identifier from the source URL
        statute_type (int, optional): [description]. Defaults to 2.
    """

    # if the path already exists, no need to create
    p = Path(".") / "tests" / "data" / f"{get_context(statute_type)}"
    if not p.exists():
        p.mkdir()

    # if the file already exists, no need to proceed
    target_file = p / f"{idx}.html"
    if target_file.exists():
        print(f"Already existing {idx}.html. No need to scrape.")
        return

    target_file.write_text(raw_from_url(idx, statute_type))


def get_content_from_file(loc: Path, idx: int, statute_type: int):
    p = loc / f"{get_context(statute_type)}"
    if not p.exists():
        print(f"Not found: {str(p)}")
        return

    # if the file already exists, no need to proceed
    target_file = p / f"{idx}.html"
    if not target_file.exists():
        print(f"Not found: {str(target_file)}")

    return target_file.read_text()


def sanitize_html_file(p: Path) -> str:
    """Get text from folder

    Args:
        p (Path): Location of the text

    Returns:
        str: The raw content
    """
    with codecs.open(str(p), "r") as f:  # codecs uses string
        return sanitize_content(f.read())


def cleaned_from_file(filename: str, context: str) -> Optional[dict]:
    """Source content from the tests / data folder to produce a data dictionary

    Args:
        filename (str): [description]

    Returns:
        [type]: [description]
    """
    p = Path(".") / "tests" / "data" / f"{filename}.html"
    if not p.exists():
        print(f"Not found: {str(p)}")
        return None
    raw = sanitize_html_file(p)
    data = parse(raw, context)
    return data



import re
from typing import Any, Match, Optional, Union

from bs4 import BeautifulSoup, NavigableString, PageElement


def pre_singular_enacter_mark(
    elem: PageElement,
) -> bool | Match | None:
    NOW_THEREFORE_MARK = re.compile(
        r"""
            NOW
            \,
            \s*
            THEREFORE
            \,
            \s*
            I\,
        """,
        re.X,
    )
    return (
        isinstance(elem.next_element, NavigableString)
        and elem.next_element is not None
        and NOW_THEREFORE_MARK.search(elem.next_element)
    )


def pre_collegiate_enacter_mark(
    elem: PageElement,
) -> bool | Match | None:
    """Get elements before the phrase 'Be it enacted', etc."""
    GENERAL_ENACTING_MARK = re.compile(
        r"""
        ^ # start
        \s*
        (
            (
                Be|
                By # mistyped 'By' (should be 'Be')
            )
            \s+
            (
                it
                \s+
            )? # includes missing 'it'
            enacted
        )
        |(
            By
            \s+
            authority
            \s+
            of
            \s+
            the
            \s+
            United
            \s+
            States
        )""",
        re.X,
    )

    return (
        isinstance(elem.next_element, NavigableString)
        and elem.next_element is not None
        and GENERAL_ENACTING_MARK.search(elem.next_element)
    ) or (
        isinstance(elem, NavigableString)
        and GENERAL_ENACTING_MARK.search(elem)
    )


def get_enacter_marker(
    soup: BeautifulSoup, classification: str
) -> PageElement | None:
    """Get the element which marks the start of the 'enacters' of the law.

    Args:
        soup (BeautifulSoup): The html markup of the whole statute
        classification (str): Flag which is either "collegiate" or "singular"

    Returns:
        Optional[PageElement]: The page element (representing the enacter) to be used by the calling function
    """
    pre_marker = None

    if classification == "collegiate":
        pre_marker = soup.find(pre_collegiate_enacter_mark)

    elif classification == "singular":
        pre_marker = soup.find(pre_singular_enacter_mark)

    else:
        return None

    return pre_marker.next_element if pre_marker else None



import re

import arrow
from bs4 import BeautifulSoup


def clean(text: str) -> str:
    translation = str.maketrans({"\n": "", "\t": ""})
    return text.translate(translation).strip()


def slice_extra(soup: BeautifulSoup) -> str:
    body = str(soup)
    title = str(soup.find("h3"))
    idx = body.find(title) + len(title)
    raw_text = body[idx:]
    return clean(raw_text)


def get_numeral(numeric_title: str) -> str:

    parts = numeric_title.strip().split()

    candidate = parts[-1]

    if candidate == "[*]":
        numeral = parts[-2]

    elif candidate == "[1]":
        numeral = parts[-2]

    elif match := re.search(r"\d+\[\*\]", candidate):
        numeral = candidate.removesuffix("[*]")

    else:
        numeral = candidate

    return numeral


def get_date(text: str) -> str:
    date_found = arrow.get(text, "MMMM DD, YYYY")
    date_format = date_found.format("MMMM DD, YYYY")
    return date_format


def get_item(title_text: str, date_text: str) -> str:
    """The first h2 element in the soup is the numeric title of the law
    e.g. [ PRESIDENTIAL DECREE NO. 1458-A, June 11, 1978 ]

    Args:
        title_text (str): The title text, see example above
        date_text (str): The formatted date string found

    Returns:
        str: The title_text without other details, i.e. Presidential Decree No. 1458-A
    """
    text = clean(title_text)
    text_without_date = re.sub(date_text, "", text)
    return text_without_date.strip("[*], ").title()


def get_title_data(text: str) -> dict:
    """Given a the header portion of the text, get the necessary fields.

    Args:
        text (str): The text containing the title elements

    Returns:
        dict: With item, law_title, date, extra keys.
    """
    soup = BeautifulSoup(text, "html5lib")
    numeric_title = soup.find("h2").get_text()
    return {
        "law_title": soup.find("h3").get_text().title(),
        "date": (date_pub := get_date(numeric_title)),
        "item": (item := get_item(numeric_title, date_pub)),
        "numeral": (item := get_numeral(item)),
        "extra": slice_extra(soup),
    }



import re
from typing import Iterable, Match, NamedTuple, Optional

from bs4 import BeautifulSoup, NavigableString, PageElement, Tag

from .sanitizer import statute_sanitizer


def slice_unit(item: Tag, source: str) -> str:
    """Slice a snippet from the source based on the item parameter as the start index, and its succeeding item as the end index

    Args:
        item (Tag): the section tag
        source (str): the html markup of the entire text

    Returns:
        str: the characters between the item parameter and the next
    """
    s: int = source.index(str(item))
    mark: str = str(end) if (end := item.find_next("section")) else "</body>"
    e: int = source.index(mark)
    return source[s:e]


class Unit(NamedTuple):
    order: int
    item: str
    caption: str
    content: str


def caption_content(raw: str):

    # initialize controls
    caption = None
    counter = 0
    soup = BeautifulSoup(raw, "html5lib")
    child_nodes = soup.body.contents

    # remove section; cannot use decompose since the section might contain relevant text
    soup.section.unwrap()

    while True:
        try:
            target = child_nodes[counter]
        except:
            break

        # first non-empty string
        if isinstance(target, NavigableString) and str(target).strip():
            break

        # italicized text, extract from child_nodes is done implicitly
        elif isinstance(target, Tag) and target.name == "em":
            caption = target.extract().get_text().strip("–—- ")
            break

        counter += 1

    content = "".join(str(i) for i in child_nodes).strip("–—- ")
    return caption, statute_sanitizer.sanitize(content)


def make_unit(elem: Tag, html_markup: str) -> Optional[Unit]:
    """Using passed section tag, generate a section snippet as a Unit

    Args:
        elem (Tag): [description]
        soup (BeautifulSoup): [description]

    Returns:
        Unit: [description]
    """
    if not (section_number := match_section(elem.get_text())):
        return None
    item = section_number.group().strip()
    uniform_item = uniform_section_phrase(item)
    sliced = slice_unit(elem, html_markup)
    removed_item = re.sub(item, "", sliced).strip()
    caption, content = caption_content(removed_item)
    return Unit(int(elem["id"]), uniform_item, caption, content)


def make_units(html: BeautifulSoup) -> Iterable[Unit]:
    """The html object will be used to generate section tags
    For each (pre-marked) section tag, create a unit

    Args:
        html (BeautifulSoup): The marked html

    Returns:
        Iterable[Unit]: named tuple consisting of unit parameters

    Yields:
        Iterator[Iterable[Unit]]: [description]
    """
    revised_html = str(html)
    for section in html("section"):
        if unit := make_unit(section, revised_html):
            yield unit


def match_section(raw: str) -> Match | None:
    """Return matching section text from the raw string, if it exists

    Args:
        raw (str): [description]

    Returns:
        Optional[Match]: [description]
    """
    regex = r"""
        ^\s*
        S
        (
            (
                (EC|ec)
                [\s\.\,]*
                \d+ # any number
                \.
                \s*
            )|
            (
                ECTION
                \s*
                1 # first section
                \.
                \s*
            )
        )
    """
    pattern = re.compile(regex, re.X)
    return pattern.search(raw)


def uniform_section_phrase(raw: str):
    regex = r"""
        ^\s*
        S
        (
            ECTION|
            EC|
            ec
        )
        [\s.,]+
    """
    pattern = re.compile(regex, re.X)
    return (
        pattern.sub("Section ", raw).strip("., ")
        if pattern.search(raw)
        else raw
    )


def candidate_unit(el: PageElement):
    """Tag matching the following requisites: (1) the next element is a navigable string,
    (2) the next element is not None, and (3) the string version of the next element matches
    the the section pattern to be be considered a candidate for a unit

    Args:
        el (PageElement): Target BeautifulSoup object to flag

    Returns:
        [type]: [description]
    """
    return (
        isinstance(el.next_element, NavigableString)
        and el.next_element is not None
        and match_section(str(el.next_element))
    )


def mark_html_sections(raw: str) -> BeautifulSoup | None:
    html = BeautifulSoup(raw, "html5lib")
    if candidates := html(candidate_unit):
        for counter, candidate in enumerate(candidates, start=1):
            wrapper = html.new_tag("section")
            target = candidate.next_element
            elem = target.wrap(wrapper)
            elem["id"] = counter
        return list(make_units(html))
    return None


def list_sections(raw: str) -> list[dict]:
    """Based on the markup in html format, generate units. For each unit generated, convert into a dictionary.

    Args:
        raw (str): [description]

    Returns:
        list[dict]: [description]
    """
    return (
        [s._asdict() for s in raw_s]
        if (raw_s := mark_html_sections(raw))
        else []
    )



from typing import Optional

from .demarcator import get_demarcated_data
from .sanitizer import statute_sanitizer
from .setter import cull_footer, set_data


def parse(raw: str, context: str) -> Optional[dict]:
    """Parsed the raw text under a given context, e.g. "ra", "pd", etc.
    The footer needs to be culled from the raw text.
    The raw text needs to be demarcated, depending on the context.
    The demarcated text needs to be split.
    The resulting data dictionary is based on the text demarcated.

    Args:
        raw (str): stringified HTML markup
        context (str): determines demarcations for slicing the raw text

    Returns:
        dict: Results of demarcated splits and culling
    """
    sanitized = statute_sanitizer.sanitize(raw)
    culled = cull_footer(sanitized)
    demarcated = get_demarcated_data(culled, context)
    data = set_data(demarcated)
    return data



import re

from bs4 import BeautifulSoup

from .elements import get_enacter_marker, get_signer_marker


def get_demarcated_data(text: str, context: str) -> dict:
    """Get three parts depending on the context.
    The context determines the classification.
    The classification determines what to use for markers of both enacter and signer clauses.

    Args:
        text (str): The full text of the law
        context (str): e.g. ra, bp, etc. Note that RAs, BPs are classified as "collegiate", EOs, PDs as "singular"

    Returns:
        dict: with the following keys
        1. raw: statute as marked
        2. enacter: the tag indicating the enacters in string format
        3. signer: the tag indicating the signers
    """

    soup = BeautifulSoup(text, "html5lib")

    flag = "singular" if context in ["pd", "eo"] else "collegiate"

    enacter_mark = get_enacter_marker(soup, flag)
    enacter_tag = None
    if enacter_mark:
        enacter_tag = soup.new_tag("p")
        enacter_tag["data-type"] = "enacter"
        enacter_mark.wrap(enacter_tag)

    signer_mark = get_signer_marker(soup, flag)
    signer_tag = None
    if signer_mark:
        signer_tag = soup.new_tag("p")
        signer_tag["data-type"] = "signer"
        signer_mark.wrap(signer_tag)

    return {
        "raw": str(soup),
        "enacter": str(enacter_tag) if enacter_tag else None,
        "signer": str(signer_tag) if signer_tag else None,
    }



import re
from typing import Any, Match, Optional, Union

import arrow
from bs4 import BeautifulSoup, NavigableString, PageElement


def lapse_date_mark(
    elem: PageElement,
) -> Union[bool, Match[Any], None]:
    LAPSE = r"Lapsed\s+into\s+law"
    return (
        isinstance(elem.next_element, NavigableString)
        and elem.next_element is not None
        and re.search(LAPSE, elem.next_element)
    )


def get_lapse_date_line(raw: str) -> Optional[str]:
    soup = BeautifulSoup(raw, "html5lib")
    mark = soup.find(lapse_date_mark)
    return str(mark.next_element) if mark else None



import re
from typing import Optional

import arrow
from arrow.arrow import Arrow
from arrow.parser import ParserMatchError
from bs4 import BeautifulSoup


def match_date(text: str) -> tuple[Optional[Arrow], Optional[str]]:
    date_found = None
    date_format = None

    try:  # with comma
        date_found = arrow.get(text, "MMMM D, YYYY")
        date_format = date_found.format("MMMM D, YYYY")

    except ParserMatchError:
        pass

    try:  # without comma
        date_found = arrow.get(text, "MMMM D YYYY")
        date_format = date_found.format("MMMM D YYYY")

    except ParserMatchError:
        pass

    return date_found, date_format


def process(raw: str):
    formatted_date = None
    item = raw.strip().title()

    date_found, original_format = match_date(raw)
    if date_found and original_format:
        formatted_date = date_found.format("MMMM D, YYYY")
        item = re.sub(original_format, "", raw)

    return {
        "publication": item.strip("(), ").title(),
        "date_of_publication": formatted_date,
    }


def parse_publications(raw: Optional[str]) -> list:
    if raw and ";" in raw:
        return [process(pub) for pub in raw.split(";")]
    elif raw and ";" not in raw:
        return [process(raw)]
    return []


def slice_publication_content(raw: str) -> str:
    """Get the first line found in the body, usually the publication string

    Args:
        raw (str): heading text,

    Returns:
        str: Possible publication string; can be an empty string since some laws do not have this filled out

    Example:
        >>> raw_html = "S. No. 61 H. No. 7090; 106 OG No. 34, 4661 (August 23, 2010); Philippine Star; Business Mirror, August 16, 2010<BR><BR><CENTER><H2 style='background-color:#cccccc;padding-top:10px;padding-bottom:10px;'>[ REPUBLIC ACT NO. 10142, July 18, 2010 ]</H2><H3>AN ACT PROVIDING FOR THE REHABILITATION OR LIQUIDATION OF FINANCIALLY DISTRESSED ENTERPRISES AND INDIVIDUALS</H3></CENTER>"
        >>> get_publications(raw_html)
        'S. No. 61 H. No. 7090; 106 OG No. 34, 4661 (August 23, 2010); Philippine Star; Business Mirror, August 16, 2010'
    """
    soup = BeautifulSoup(raw, "html5lib")

    # create a marker for slicing
    center_elem = soup.find("center")
    center_elem["id"] = "title-law"
    target_markup = str(center_elem)

    # convert soup to string
    html_markup = str(soup)

    # get the start index
    pre = "E-Library - Information At Your Fingertips: Printer Friendly"
    match = re.search(pre, html_markup)
    s = match.end() if match else 0

    # get the end index
    e = html_markup.index(target_markup)

    # get the slice based on the index
    sliced_text = html_markup[s:e]

    # convert slice back to soup and strip; note this can be empty
    html = BeautifulSoup(sliced_text, "html5lib")
    text = html.get_text().strip()

    return text


def get_publication_data(raw: str):
    text = slice_publication_content(raw)
    return parse_publications(text)



import re
from typing import Any, Match, Optional, Union

from bs4 import BeautifulSoup, NavigableString, PageElement


def pre_singular_signer_mark(
    elem: PageElement,
) -> Union[bool, Match[Any], None]:
    """Get elements before the phrase 'Done in the City of Manila,'

    Args:
        elem (PageElement): bs4 required element placeholder, see docs

    Returns:
        bool: True for all elements matching the criteria
    """
    MANILA_INDICATOR = r"Done\s+in\s+the\s+City\s+of\s+Manila\,"
    return (
        isinstance(elem.next_element, NavigableString)
        and elem.next_element is not None
        and re.search(MANILA_INDICATOR, elem.next_element)
    )


def pre_collegiate_signer_mark(
    elem: PageElement,
) -> Union[bool, Match[Any], None]:
    """Get elements before the phrase 'Approved,'

    Args:
        elem (PageElement): bs4 required element placeholder, see docs

    Returns:
        bool: True for all elements matching the criteria
    """
    APPROVED_INDICATOR = r"Approved\,"
    return (
        isinstance(elem.next_element, NavigableString)
        and elem.next_element is not None
        and re.search(APPROVED_INDICATOR, elem.next_element)
    )


def get_signer_marker(
    soup: BeautifulSoup, classification: str
) -> Optional[PageElement]:
    """Get the element which marks the start of the 'signers' of the law.

    Args:
        soup (BeautifulSoup): The html markup of the whole statute
        classification (str): Flag which is either "collegiate" or "singular"

    Returns:
        Optional[PageElement]: The page element (representing the signer) to be used by the calling function
    """
    pre_marker = None

    if classification == "collegiate":
        pre_marker = soup.find(pre_collegiate_signer_mark)

    elif classification == "singular":
        pre_marker = soup.find(pre_singular_signer_mark)

    else:
        return None

    return pre_marker.next_element if pre_marker else None



from .enacters import get_enacter_marker
from .lapse import get_lapse_date_line
from .publications import get_publication_data
from .signers import get_signer_marker
from .titles import get_numeral, get_title_data



from typing import Optional

from bs4 import BeautifulSoup, Tag

from .elements import get_lapse_date_line, get_publication_data, get_title_data
from .sections import list_sections


def cull_footer(text: str) -> str:
    """Remove dividing text that are unnecessary for parsing the law

    Args:
        text (str): the full content

    Returns:
        str: the stringified html markup with the unnnecessary parts removed
    """
    soup = BeautifulSoup(text, "html5lib")
    for ender in soup("hr"):  # see cases like 86448
        [i.decompose() for i in ender.next_siblings if isinstance(i, Tag)]
        ender.decompose()

    return str(soup)


def set_data(data: dict) -> Optional[dict]:
    """Presumes previous demarcation function has been applied, resulting in a data dictionary.
    The data dictionary can be used in the `acquisition` library.

    Args:
        data (dict): Consists of demarcations for: (1) raw text, (2) enacter text, and (3) signer text.

    Returns:
        Optional[dict]: If processed properly, a dictionary of fields.
    """

    # initialize
    content = data["raw"]
    header_portion = None
    enacting_clause = None
    signing_clause = None
    lapse_date_line = None
    units = []

    # since content has been demarcated, get the enacter phrase, if it exists
    # the text until the enacter phase is the header portion
    if enacter := data["enacter"]:
        enacter_start_idx = content.find(enacter)
        enacter_end_idx = enacter_start_idx + len(enacter)
        header_portion = content[:enacter_start_idx]
        enacting_clause = content[enacter_start_idx:enacter_end_idx]

    # since content has been demarcated, get the signer phrase, if it exists
    # the text may include the "lapse into law line"
    if signer := data["signer"]:
        signer_idx = content.find(signer)
        signing_clause = content[signer_idx:]
        lapse_date_line = get_lapse_date_line(signing_clause)

    # if the main clauses are found in the content, can remove these to get the gist
    # the gist is necessary to secure unit "sections"
    if enacting_clause and signing_clause:
        start = content.find(enacting_clause) + len(enacting_clause)
        end = content.find(signing_clause)
        content = content[start:end]
        units = list_sections(content)

    if not header_portion:
        return None

    return {
        "publications": get_publication_data(header_portion),
        "statute_gist": content,
        "enacting_clause": enacting_clause,
        "signers_of_law": signing_clause,
        "lapse_into_law_clause": lapse_date_line,
        "units": units,
    } | get_title_data(header_portion)
